﻿cls;
#Define our connection string properties
$ConnectionString = "Data Source=.;Initial Catalog=TestImplicitConvert_WinCollation;Integrated Security=True";

#Create a connection to SQL Server
$SqlConn = New-Object System.Data.SqlClient.SqlConnection;
$SqlConn.ConnectionString = $ConnectionString;
$SqlConn.Open();

#Collect the column names from our table into an array
$ColList = New-Object System.Collections.ArrayList;

$SqlCmd = $SqlConn.CreateCommand();
$SqlCmd.CommandText = "	SELECT c.name FROM sys.columns AS c WHERE object_id = OBJECT_ID('TestImplicitConverts') AND c.name <> 'RowID';"
$SqlCmd.CommandType = [System.Data.CommandType]::Text;
$SqlReader = $SqlCmd.ExecuteReader();

while ($SqlReader.Read())
{
	$ColList.Add($SqlReader.GetSqlString(0)) | Out-Null;
}

$SqlReader.Close();
$SqlConn.Close();

#Iterate each column name and query the table for that column using each data type
foreach($col in $ColList)
{
	#Get the data types we wish to test, excluding less common types
	[Enum]::GetValues([System.Data.SqlDbType]) | `
		WHERE {
				$_ -ne "Binary" -and 
				$_ -ne "Image" -and 
				$_ -ne "NText" -and 
				$_ -ne "Text" -and 
				$_ -ne "Timestamp" -and
				$_ -ne "VarBinary" -and
				$_ -ne "Variant" -and
				$_ -ne "Xml" -and
				$_ -ne "Udt" -and
				$_ -ne "Structured"
				} | % {

		$type = $_;		
		$typename = $type.ToString();
		
		#Set the parameter value based on the type being tested
		switch ($typename)
		{
			"BigInt" {$val = 1}
			"Bit" {$val = 1}
			"Char" {$val = "1"}
			"Decimal" {$val = 1}
			"Float" {$val = 1}
			"Int" {$val = 1}
			"Money" {$val = 1}
			"NChar" {$val = "1"}
			"NVarChar" {$val = "1"}
			"Real" {$val = 1}
			"UniqueIdentifier"  {$val = [Guid]::NewGuid()}
			"SmallDateTime" {$val = [DateTime]::Now}
			"SmallInt" {$val = 1}
			"SmallMoney" {$val = 1}
			"TinyInt" {$val = 1}
			"VarChar" {$val = "1"}
			"Date" {$val = [DateTime]::Now}
			"DateTime" {$val = [DateTime]::Now}
			"DateTime2" {$val = [DateTime]::Now}
			"DateTimeOffset" {$val = [DateTime]::Now}
			default {$val = "0"}			
		}
		
		#Connect to SQL Server
		$SqlConn = New-Object System.Data.SqlClient.SqlConnection;
		$SqlConn.ConnectionString = $ConnectionString;
		$SqlConn.Open();
		
		#Turn on our ShowplanXML collection
		$SqlCmd = New-Object System.Data.SqlClient.SqlCommand;
		$SqlCmd.Connection = $SqlConn;
		$SqlCmd.CommandText = "SET STATISTICS XML ON";
		
		#Query the column on our table filtering the column on our current type and value
		$SqlCmd2 = New-Object System.Data.SqlClient.SqlCommand;
		$SqlCmd2.Connection = $SqlConn;
		$SqlCmd2.CommandText = [string]::Format("SELECT {0} FROM TestImplicitConverts WHERE {1} = @Param;", $col, $col);
		$SqlCmd2.CommandType = [System.Data.CommandType]::Text;
		
		#Create our parameter and assign it our type and value
		$SqlParameter = $SqlCmd2.Parameters.AddWithValue("@Param", $val);
		$SqlParameter.SqlDbType = $type;
		
		#Turn off the ShowplanXML collection
		$SqlCmd3 = New-Object System.Data.SqlClient.SqlCommand;
		$SqlCmd3.Connection = $SqlConn;
		$SqlCmd3.CommandText = "SET STATISTICS XML OFF";
		
		#Clear variables before running our test
		$SqlException = $null;
		$String = [string]::Empty;
		
		try {
			#Turn on the ShowplanXML 
			$SqlCmd.ExecuteNonQuery() | Out-Null;
			
			#Start our data reader
			$SqlDR = $SqlCmd2.ExecuteReader();
		
			## skip over returned results
			while ($SqlDR.Read())
			{ } 
			
			## Move to next result which has the ShowplanXML
			$SqlDR.NextResult() | Out-Null;
			if ($SqlDR.Read())
			{
				$String = $SqlDR[0].ToString();
			}
			$SqlDR.Close();
			
			#Turn off the ShowplanXML
			$SqlCmd3.ExecuteNonQuery() | Out-Null;

			#Close the connection
			$SqlConn.Close();
			
			#Load the Showplan as a XML document for parsing
			$xml = New-Object System.Xml.XmlDocument;
			$xml.LoadXml($String);
			
			#Setup the XmlNamespaceManager and add the ShowPlan Namespace to it.		
			$nsMgr = New-Object 'System.Xml.XmlNamespaceManager' $xml.NameTable;
			$nsMgr.AddNamespace("sm", "http://schemas.microsoft.com/sqlserver/2004/07/showplan");
			
			#Search for a Index Scan or Clustered Index Scan operation that is an Implicit Conversion
			$HasImplictConvert = $xml.SelectNodes("//sm:RelOp[@LogicalOp='Index Scan' or @LogicalOp='Clustered Index Scan']/sm:IndexScan/sm:Predicate//sm:Convert[@Implicit='1']/sm:ScalarOperator/sm:Identifier/sm:ColumnReference", $nsMgr).Count -gt 0;
		}
		catch 
		{
			#Handle any errors
			$err = $_;	
			if ($_.Exception.InnerException.GetType() -eq [System.Data.SqlClient.SqlException])
			{	$SqlException = [System.Data.SqlClient.SqlException]$_.Exception.InnerException;	}
			else
			{
				Write-Host $_.Exception;
			}
		}
		finally
		{
			#Perform final clean up
			if($SqlDR.IsClosed -eq $false)
			{	$SqlDR.Close();	}
			
			if($SqlConn.State -eq [System.Data.ConnectionState]::Open)
			{	$SqlConn.Close();	}
			
			#Output the results in CSV to the window for usage.
			$msg = [string]::Format("{0}, {1}, {2}, {3}, {4}", $col, $type, $HasImplictConvert, $SqlException.Number, $SqlException.Message);
			Write-Host $msg;
		}
	}
}